mod model;

use std::{collections::HashMap, fs::File, io::BufReader, path::PathBuf, sync::OnceLock};

use path_macro::path;

use model::{GenreEntry, GenreMeta};

pub use model::{Genre, Region};

type Translation = HashMap<String, String>;

pub struct GenreDb {
    /// Path to genre package
    path: PathBuf,
    /// Genre database
    genres: HashMap<String, GenreEntry>,
    /// Map of lazily loaded translations by language
    translations: HashMap<String, OnceLock<Option<Translation>>>,
}

#[derive(thiserror::Error, Debug)]
pub enum Error {
    #[error("{0}")]
    File(#[from] std::io::Error),
    #[error("{0}")]
    Serde(#[from] serde_json::Error),
}

impl GenreDb {
    /// Load a new [`GenreDb`] from the JSON package at the given location
    pub fn new<P: Into<PathBuf>>(path: P) -> Result<Self, Error> {
        let path: PathBuf = path.into();

        // Find available languages
        let mut translations = HashMap::new();
        for entry in std::fs::read_dir(path!(path / "lang"))? {
            let entry = entry?;
            if let Some(lang) = entry
                .file_name()
                .to_str()
                .and_then(|n| n.strip_suffix(".json"))
            {
                translations.insert(lang.to_owned(), OnceLock::new());
            }
        }

        let db_path = path!(path / "genres.json");
        let db_file = File::open(db_path)?;

        Ok(Self {
            path,
            genres: serde_json::from_reader(BufReader::new(db_file))?,
            translations,
        })
    }

    /// Return a list of supported languages
    pub fn languages(&self) -> Vec<&str> {
        self.translations.keys().map(String::as_str).collect()
    }

    /// Select a supported language from a list of given languages (for example from the
    /// Accept-Language http headere).
    pub fn select_lang<'a, S: AsRef<str>>(&self, langs: &'a [S]) -> Option<&'a str> {
        for lang in langs {
            let lang = lang.as_ref();
            if self.translations.contains_key(lang) {
                return Some(lang);
            }
            if lang == "en" || lang.starts_with("en-") {
                return None;
            }
        }
        None
    }

    pub fn get_name(&self, id: &str) -> Option<String> {
        self.get_localized_name(id, None)
    }

    /// Return the name of a genre in the given langage
    pub fn get_localized_name(&self, id: &str, lang: Option<&str>) -> Option<String> {
        match self.genres.get(id)? {
            GenreEntry::Alias { alias } => self.get_localized_name(alias, lang),
            GenreEntry::Meta(genre) => {
                if let Some(lang) = lang {
                    if let Some(tn) = self.get_translated_name(id, lang) {
                        return Some(tn);
                    }
                }
                Some(genre.name.to_owned())
            }
        }
    }

    /// Return the name of a genre from a translation file
    ///
    /// Note: this function does not fall back to the default (English) translation
    fn get_translated_name(&self, id: &str, lang: &str) -> Option<String> {
        let translation = self.translations.get(lang)?;
        let tl = translation.get_or_init(|| {
            let lang_path = path!(self.path / "lang" / format!("{lang}.json"));
            let lang_file = match File::open(lang_path) {
                Ok(file) => file,
                Err(e) => {
                    log::error!("could not open genre language file `{lang}`: {e}");
                    return None;
                }
            };
            match serde_json::from_reader(BufReader::new(lang_file)) {
                Ok(res) => Some(res),
                Err(e) => {
                    log::error!("could not deserialize genre language file `{lang}`: {e}");
                    None
                }
            }
        });
        tl.as_ref().and_then(|tl| tl.get(id).cloned())
    }

    /// Get a genre (with the English name)
    pub fn get(&self, id: &str) -> Option<Genre> {
        self.get_localized(id, None)
    }

    /// Get a genre (with the localized name)
    pub fn get_localized(&self, id: &str, lang: Option<&str>) -> Option<Genre> {
        match self.genres.get(id)? {
            GenreEntry::Alias { alias } => self.get_localized(alias, lang),
            GenreEntry::Meta(genre) => {
                if let Some(lang) = lang {
                    Some(
                        self.conv_genre(
                            id,
                            self.get_translated_name(id, lang)
                                .unwrap_or_else(|| genre.name.to_owned()),
                            genre,
                        ),
                    )
                } else {
                    Some(self.conv_genre(id, genre.name.to_owned(), genre))
                }
            }
        }
    }

    fn conv_genre<'a>(&self, id: &str, name: String, gm: &GenreMeta) -> Genre {
        let mut genre = Genre {
            id: id.to_owned(),
            name: name,
            parent: gm.parent.to_owned(),
            language: gm.language.to_owned(),
            country: gm.country.to_owned(),
            region: gm.region,
            rank: gm.rank,
            deprecated: gm.deprecated,
            metagenre: gm.metagenre,
            #[cfg(feature = "playlists")]
            playlists: gm.playlists.clone(),
        };
        self.inherit_parent(&mut genre, gm.parent.as_deref());
        genre
    }

    fn inherit_parent(&self, genre: &mut Genre, parent: Option<&str>) {
        if let Some(GenreEntry::Meta(g2)) = parent.and_then(|p| self.genres.get(p)) {
            merge_genre_data(genre, g2);
            self.inherit_parent(genre, g2.parent.as_deref());
        }
    }

    /// Get the path of the genre tree
    pub fn tree_path(&self, lang: Option<&str>) -> PathBuf {
        match lang.filter(|l| self.translations.contains_key(*l)) {
            Some(lang) => path!(self.path / "tree" / format!("tree.{lang}.json")),
            None => path!(self.path / "tree" / "tree.en.json"),
        }
    }
}

fn merge_genre_data(genre: &mut Genre, g2: &GenreMeta) {
    if genre.language.is_none() {
        genre.language = g2.language.clone();
    }
    if genre.country.is_none() {
        genre.country = g2.country.clone();
    }
    if genre.region.is_none() {
        genre.region = g2.region.clone();
    }
}

#[cfg(test)]
mod tests {
    use super::*;

    fn sync_and_send<T: Sync + Send>(_x: T) {}

    fn genre_db_path() -> PathBuf {
        path!(env!("CARGO_MANIFEST_DIR") / ".." / ".." / "dist")
    }

    #[test]
    fn get_localized() {
        let db = GenreDb::new(genre_db_path()).unwrap();
        sync_and_send(&db);

        let id = "albanian rock";
        let lang = Some("de");
        let expect = "Albanischer Rock";

        let genre = db.get_localized(id, lang).unwrap();
        assert_eq!(genre.id, id);
        assert_eq!(genre.name, expect);
        assert_eq!(genre.language.unwrap(), "sqi");
        assert_eq!(genre.country.unwrap(), "AL");

        let localized_name = db.get_localized_name(id, lang).unwrap();
        assert_eq!(localized_name, expect)
    }

    #[test]
    /// Genres should inherit parent parameters
    fn get_subgenre() {
        let db = GenreDb::new(genre_db_path()).unwrap();
        let genre = db.get("k-pop girl group").unwrap();
        assert_eq!(genre.language.unwrap(), "kor");
        assert_eq!(genre.country.unwrap(), "KR");
    }

    #[test]
    fn select_lang() {
        let languages = vec!["xy".to_owned(), "de".to_owned(), "en-GB".to_owned()];

        let db = GenreDb::new(genre_db_path()).unwrap();
        let lang = db.select_lang(&languages);
        assert_eq!(lang, Some("de"));
    }

    #[test]
    fn select_lang_en() {
        let languages = vec!["en-GB".to_owned(), "de".to_owned()];

        let db = GenreDb::new(genre_db_path()).unwrap();
        let lang = db.select_lang(&languages);
        assert_eq!(lang, None);
    }

    #[test]
    fn tree_path() {
        let db_path = genre_db_path();
        let db = GenreDb::new(&db_path).unwrap();

        let expect_en = path!(&db_path / "tree" / "tree.en.json");
        let expect_de = path!(&db_path / "tree" / "tree.de.json");

        assert_eq!(db.tree_path(None), expect_en);
        assert_eq!(db.tree_path(Some("en")), expect_en);
        assert_eq!(db.tree_path(Some("xy")), expect_en);
        assert_eq!(db.tree_path(Some("de")), expect_de);
    }
}
