#[cfg(feature = "playlists")]
use std::collections::HashMap;

use serde::{Deserialize, Serialize};
use serde_with::{serde_as, DefaultOnError};

/// Region for describing genres from multiple countrie
#[derive(Debug, Copy, Clone, PartialEq, Eq, Hash, Serialize, Deserialize)]
#[non_exhaustive]
pub enum Region {
    /// Africa
    AF,
    /// Balkan
    BK,
    /// Baltic
    BT,
    /// Central America
    CA,
    /// Caribbean
    CB,
    /// Caucasian
    CC,
    /// Celtic
    CE,
    /// Canada+Greenland
    CG,
    /// Czech+Slovakia
    CK,
    /// Central Asia
    CS,
    /// D-A-CH
    DH,
    /// Desi
    DS,
    /// Eastern Europe
    EE,
    /// Europe
    EU,
    /// Hawaii
    HW,
    /// Latin America
    LA,
    /// Mediterranean
    ME,
    /// Nordic
    ND,
    /// North Africa
    NF,
    /// Oceania
    OZ,
    /// South America
    SA,
    /// Scandinavia
    SC,
    /// Southeast Asia
    SE,
    /// West Africa
    WF,
}

serde_plain::derive_display_from_serialize!(Region);

#[derive(Debug, Clone, Serialize, Deserialize)]
#[serde(untagged)]
pub enum GenreEntry {
    Alias { alias: String },
    Meta(GenreMeta),
}

/// Genre metadata
#[serde_as]
#[derive(Debug, Clone, Serialize, Deserialize)]
pub struct GenreMeta {
    pub name: String,
    pub parent: Option<String>,
    pub language: Option<String>,
    pub country: Option<String>,
    #[serde(default)]
    #[serde_as(deserialize_as = "DefaultOnError")]
    pub region: Option<Region>,
    pub rank: Option<u32>,
    #[cfg(feature = "playlists")]
    #[serde(default)]
    pub playlists: HashMap<String, String>,
    #[serde(default)]
    pub deprecated: bool,
    #[serde(default)]
    pub metagenre: bool,
}

#[serde_as]
#[derive(Debug, Clone, Serialize, Deserialize)]
#[non_exhaustive]
pub struct Genre {
    /// Spotify genre ID
    pub id: String,
    /// Localized genre name
    pub name: String,
    /// ID of the parent genre if this is a subgenre
    #[serde(skip_serializing_if = "Option::is_none")]
    pub parent: Option<String>,
    /// ISO-639-3 language code if the genre implies lyrics in a specific language
    #[serde(skip_serializing_if = "Option::is_none")]
    pub language: Option<String>,
    /// ISO-3166-1 country code if the genre is dominant in a specific country
    #[serde(skip_serializing_if = "Option::is_none")]
    pub country: Option<String>,
    /// Region code if the genre is dominant in a specific region
    ///
    /// Region codes are not standardized, refer to the documentation for a definition.
    #[serde(default, skip_serializing_if = "Option::is_none")]
    #[serde_as(deserialize_as = "DefaultOnError")]
    pub region: Option<Region>,
    /// Position in the popularity ranking (1: most popular)
    #[serde(skip_serializing_if = "Option::is_none")]
    pub rank: Option<u32>,
    /// Spotify playlist IDs
    #[cfg(feature = "playlists")]
    #[serde(skip_serializing_if = "Option::is_none")]
    pub playlists: Option<HashMap<String, String>>,
    /// True if the genre is no longer part of Spotify's catalog
    #[serde(default, skip_serializing_if = "is_default")]
    pub deprecated: bool,
    /// True if the genre is a metagenre
    ///
    /// Metagenres do not exist at Spotify but are used to group other genres together.
    #[serde(default, skip_serializing_if = "is_default")]
    pub metagenre: bool,
}

impl PartialEq for Genre {
    fn eq(&self, other: &Self) -> bool {
        self.id == other.id
    }
}
impl Eq for Genre {}

fn is_default<T: Default + PartialEq>(t: &T) -> bool {
    t == &T::default()
}
